import oauthadmin.views

try:
    from django.urls import re_path as url
except ImportError:
    from django.conf.urls import url


urlpatterns = [
    url(r"login/", oauthadmin.views.login),
    url(r"callback/", oauthadmin.views.callback),
    url(r"logout/", oauthadmin.views.logout),
    url(r"logout_redirect/", oauthadmin.views.logout_redirect),
]



from mock import Mock
import json
from freezegun import freeze_time
from django.contrib.auth.models import AnonymousUser, User
from django.test.utils import override_settings
from oauthadmin.middleware import OauthAdminSessionMiddleware

def setup_module(mod):
   global mw
   global request
   mw = OauthAdminSessionMiddleware(Mock())
   request = Mock()
   request.session = {}

def teardown_module(mod):
   global mw
   global request
   del mw
   del request

# Note: We are not actually serializing the User
# as this required adding a database Engine.
MOCK_SERIALIZED_USER_DATA = json.dumps([{"model": "auth.user", "pk": 1, "fields": {}}])

def test_process_request_without_user():
    request.session = {}
    assert mw.process_request(request) is None
    assert isinstance(request.user, AnonymousUser)

def test_process_request_with_user():
    request.session = {'user': MOCK_SERIALIZED_USER_DATA}
    assert mw.process_request(request) is None
    assert isinstance(request.user, User)
    assert request.user.id == 1
    assert request._cached_user.id == 1


false_mock_pinger = Mock(return_value = False)

@override_settings(OAUTHADMIN_PING_INTERVAL=5)
@override_settings(OAUTHADMIN_PING='test.test_middleware.false_mock_pinger')
def test_that_anonymoususer_goes_in_request_user_if_ping_fails():
    request.user = User(id=1) # non anonymous user
    request.session = {
        'user': MOCK_SERIALIZED_USER_DATA, # non anonymous user
        'oauth_token':'abc'
    }
    mw.process_request(request)
    assert isinstance(request.user, AnonymousUser)


mock_pinger = Mock()

@override_settings(OAUTHADMIN_PING_INTERVAL=5)
@override_settings(OAUTHADMIN_PING='test.test_middleware.mock_pinger')
def test_process_request_with_reverify_interval():
    request.session = {'oauth_token': 1234, 'user': MOCK_SERIALIZED_USER_DATA}
    with freeze_time('2012-08-29 00:00:00'):
        # first request should set timestamp
        # and verify with pinger.
        mw.process_request(request)
        assert mock_pinger.called
        mock_pinger.called = False
    with freeze_time('2012-08-29 00:00:01'):
        # second request should not call pinger
        mw.process_request(request)
        assert not mock_pinger.called
    with freeze_time('2012-08-29 00:01:01'):
        # third request, after the ping interval
        # has passed, should call pinger.
        mw.process_request(request)
        assert mock_pinger.called



from django.contrib.admin.sites import AdminSite

from oauthadmin.views import login, logout_redirect

class OauthAdminSite(AdminSite):

    def login(self, request, extra_context=None):
        return login(request)

    def logout(self, request, extra_context=None):
        return logout_redirect(request)



from time import time
import base64
import json

from requests_oauthlib import OAuth2Session
from oauthlib.oauth2.rfc6749.errors import MismatchingStateError, InvalidGrantError
from oauthlib.common import generate_token
try:
    from urllib import quote_plus
except ImportError:
    from urllib.parse import quote_plus

from django.shortcuts import redirect
from django.http import HttpResponseRedirect
from django.core import serializers

from oauthadmin.utils import import_by_path
from oauthadmin.settings import app_setting
import oauthadmin.views

try:
    from django.urls import reverse, NoReverseMatch
except ImportError:
    from django.core.urlresolvers import reverse, NoReverseMatch



def destroy_session(request):

    # These session variables MAY not exist at this point.
    for key in ['oauth_state', 'oauth_token', 'uid', 'user','last_verified_at']:
        try:
            del request.session[key]
        except KeyError:
            pass

def login(request):
    # this view can be called directly by django admin site from
    # any url, or can be accessed by the login url if the urls
    # from this app were included
    if request.path == reverse(oauthadmin.views.login):
        # if this view is being accessed from login url look for 'next'
        # in query string to use as destination after the login is complete
        next = request.GET.get('next')
    else:
        # otherwise the django admin site called this view from another view.
        # Django admin doesn't redirect to login url if login is required, it
        # calls the view directly (django 1.7 fixed this and redirects and we
        # don't support it yet)
        next = request.get_full_path()

    redirect_uri = request.build_absolute_uri(reverse(oauthadmin.views.callback))
    state_token = generate_token()
    state=base64.b64encode(json.dumps({"state": state_token, "next": next}).encode('utf-8'))
    oauth = OAuth2Session(
        client_id=app_setting('CLIENT_ID'),
        redirect_uri=redirect_uri,
        scope=app_setting('SCOPE'),
        state=state,
    )
    authorization_url, state = oauth.authorization_url(app_setting('AUTH_URL'))

    request.session['oauth_state'] = state.decode('utf-8')

    return redirect(authorization_url)


def callback(request):
    if 'oauth_state' not in request.session:
        return HttpResponseRedirect(request.build_absolute_uri(reverse(oauthadmin.views.login)))
    redirect_uri = request.build_absolute_uri(reverse(oauthadmin.views.callback))
    oauth = OAuth2Session(
        app_setting('CLIENT_ID'),
        state=request.session['oauth_state'], # decoded string
        redirect_uri=redirect_uri,
    )
    try:
        token = oauth.fetch_token(
            app_setting('TOKEN_URL'),
            client_secret=app_setting('CLIENT_SECRET'),
            authorization_response=app_setting('AUTH_URL') + "?" + request.GET.urlencode()
        )
    except (MismatchingStateError, InvalidGrantError):
        return HttpResponseRedirect(request.build_absolute_uri(reverse(oauthadmin.views.login)))

    user = import_by_path(app_setting('GET_USER'))(token)
    serialized_user = serializers.serialize("json", [user])

    request.session['last_verified_at'] = int(time())
    request.session['oauth_token'] = token
    request.session['user'] = serialized_user

    next = json.loads(base64.b64decode(request.session['oauth_state']).decode('utf-8'))['next']
    if not next:
        next = app_setting('DEFAULT_NEXT_URL')

    return redirect(request.build_absolute_uri(next))


def logout(request):
    if 'oauth_token' in request.session:
        oauth = OAuth2Session(app_setting('CLIENT_ID'), token=request.session['oauth_token'])
        oauth.get(app_setting('BASE_URL') + 'destroy_tokens')

        destroy_session(request)

    return redirect(request.build_absolute_uri('/'))


def logout_redirect(request):
    return redirect(app_setting('BASE_URL') + 'logout?next=' + quote_plus(request.build_absolute_uri(reverse(oauthadmin.views.logout))))



import mock
import pytest
import base64
import json
from oauthadmin.views import destroy_session, login, callback, logout
from oauthlib.oauth2.rfc6749.errors import MismatchingStateError, InvalidGrantError
from django.test.client import RequestFactory
from django.contrib.auth.models import User
import oauthadmin.views

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


SESSION_VARIABLES = ['oauth_state', 'oauth_token', 'uid', 'user']


def _state(token, next=None):
    return base64.b64encode(json.dumps({'state': token, 'next': next or ''}).encode('utf-8'))

@pytest.fixture
def request_factory():
    return RequestFactory()

def test_destroy_session(request_factory):
    request = request_factory.get('/')
    request.session = dict([(key, key) for key in SESSION_VARIABLES])
    assert destroy_session(request) == None
    assert isinstance(request.session, dict)
    assert request.session == {}

def test_destroy_session_with_extra_values(request_factory):
    request = request_factory.get('/')
    request.session = dict([(key, key) for key in SESSION_VARIABLES] + [('extra', 'extra')])
    assert destroy_session(request) == None
    assert isinstance(request.session, dict)
    assert request.session == {'extra': 'extra'}

def test_destroy_session_with_empty_values(request_factory):
    request = request_factory.get('/')
    request.session = {}
    assert destroy_session(request) == None
    assert isinstance(request.session, dict)
    assert request.session == {}


@mock.patch('oauthadmin.views.OAuth2Session')
@mock.patch('oauthadmin.views.app_setting')
def test_login(app_setting, OAuth2Session, request_factory):
    OAuth2Session.return_value = mock.Mock(
        authorization_url = mock.Mock(return_value = ('https://foo', _state('state-variable')))
    )
    request = request_factory.get(reverse(oauthadmin.views.login))
    request.session = {}
    request.build_absolute_uri = mock.Mock(return_value='https://test.com/construct-redirect')

    app_setting.return_value = 'app-setting'

    resp = login(request)
    assert resp.status_code == 302
    assert resp['location'] == 'https://foo'
    assert request.session.get('oauth_state') == _state('state-variable').decode('utf-8')

@mock.patch('oauthadmin.views.OAuth2Session')
def test_login_redirect_uri(OAuth2Session, request_factory):
    OAuth2Session.return_value = mock.Mock(
        authorization_url = mock.Mock(return_value = ('https://foo', _state('state-variable')))
    )
    request = request_factory.get(reverse(oauthadmin.views.login))
    request.session = {}
    request.build_absolute_uri = mock.Mock(return_value='https://test.com/construct-redirect')

    resp = login(request)

    OAuth2Session.assert_called_once_with(
        client_id = 'test-client-id',
        redirect_uri = u'https://test.com/construct-redirect',
        scope = ['default'],
        state = mock.ANY,
    )

@mock.patch('oauthadmin.views.OAuth2Session')
def test_login_redirect_uri_with_next_from_url(OAuth2Session, request_factory):
    OAuth2Session.return_value = mock.Mock(
        authorization_url = mock.Mock(return_value = ('https://foo', _state('state-variable')))
    )
    request = request_factory.get(reverse(oauthadmin.views.login) + '?next=/admin/content/')
    request.session = {}
    request.build_absolute_uri = mock.Mock(return_value='https://test.com/construct-redirect')

    resp = login(request)

    OAuth2Session.assert_called_once_with(
        redirect_uri = u'https://test.com/construct-redirect',
        client_id = mock.ANY,
        scope = mock.ANY,
        state = mock.ANY,
    )

@mock.patch('oauthadmin.views.OAuth2Session')
def test_login_redirect_uri_with_next_as_current_url(OAuth2Session, request_factory):
    OAuth2Session.return_value = mock.Mock(
        authorization_url = mock.Mock(return_value = ('https://foo', _state('state-variable')))
    )
    request = request_factory.get('/admin/content/')
    request.session = {}
    request.build_absolute_uri = mock.Mock(return_value='https://test.com/construct-redirect')

    resp = login(request)

    OAuth2Session.assert_called_once_with(
        redirect_uri = u'https://test.com/construct-redirect',
        client_id = mock.ANY,
        scope = mock.ANY,
        state = mock.ANY,
    )


@mock.patch('oauthadmin.views.OAuth2Session')
@mock.patch('oauthadmin.views.app_setting')
@mock.patch('oauthadmin.views.import_by_path')
def test_callback_with_mismatching_state(import_by_path, app_setting, OAuth2Session, request_factory):
    request = request_factory.get('/')
    request.session = {'oauth_state':_state('foo')}
    app_setting.return_value = 'app-setting'
    OAuth2Session.return_value = mock.Mock(fetch_token = mock.Mock(side_effect=MismatchingStateError))
    resp = callback(request)
    assert resp.status_code == 302
    assert resp['location'] == 'http://testserver/login/'

@mock.patch('oauthadmin.views.OAuth2Session')
@mock.patch('oauthadmin.views.app_setting')
@mock.patch('oauthadmin.views.import_by_path')
def test_callback_with_missing_state(import_by_path, app_setting, OAuth2Session, request_factory):
    request = request_factory.get('/')
    request.session = {}
    app_setting.return_value = 'app-setting'
    resp = callback(request)
    assert resp.status_code == 302
    assert resp['location'] == 'http://testserver/login/'

@mock.patch('oauthadmin.views.OAuth2Session')
@mock.patch('oauthadmin.views.app_setting')
@mock.patch('oauthadmin.views.import_by_path')
def test_callback_with_invalid_grant(import_by_path, app_setting, OAuth2Session, request_factory):
    request = request_factory.get('/')
    request.session = {'oauth_state':_state('foo')}
    app_setting.return_value = 'app-setting'
    OAuth2Session.return_value = mock.Mock(fetch_token = mock.Mock(side_effect=InvalidGrantError))
    resp = callback(request)
    assert resp.status_code == 302
    assert resp['location'] == 'http://testserver/login/'

@mock.patch('oauthadmin.views.OAuth2Session')
@mock.patch('oauthadmin.views.app_setting')
@mock.patch('oauthadmin.views.serializers.serialize')
@mock.patch('oauthadmin.views.import_by_path')
def test_callback(import_by_path, mock_serialized_json_user, app_setting, OAuth2Session, request_factory):
    request = request_factory.get(reverse(oauthadmin.views.callback))
    request.session = {'oauth_state': _state('state-variable')}
    OAuth2Session.return_value = mock.Mock(
        fetch_token = mock.Mock(return_value = 'token')
    )
    app_setting.return_value = 'app-setting'
    ibp = mock.Mock()
    ibp.return_value = User(id=1)
    import_by_path.return_value = ibp

    serialized_user = '[{"model": "auth.user", "pk": 1, "fields": {}}]'
    mock_serialized_json_user.return_value = serialized_user

    resp = callback(request)
    assert resp.status_code == 302
    assert resp['location'] == 'http://testserver/callback/app-setting'
    assert request.session.get('oauth_token') == 'token'
    mock_serialized_json_user.assert_called_once_with("json", [User(id=1)])
    assert request.session.get('user') == serialized_user

@mock.patch('oauthadmin.views.OAuth2Session')
@mock.patch('oauthadmin.views.app_setting')
@mock.patch('oauthadmin.views.serializers.serialize')
@mock.patch('oauthadmin.views.import_by_path')
def test_callback_redirect_to_next(import_by_path, mock_serialized_json_user, app_setting, OAuth2Session, request_factory):
    request = request_factory.get(reverse(oauthadmin.views.callback))
    request.session = {'oauth_state': _state('state-variable','/admin/content/')}
    OAuth2Session.return_value = mock.Mock(
        fetch_token = mock.Mock(return_value = 'token')
    )
    app_setting.return_value = 'app-setting'
    ibp = mock.Mock()
    ibp.return_value = User(id=1)
    import_by_path.return_value = ibp

    serialized_user = '[{"model": "auth.user", "pk": 1, "fields": {}}]'
    mock_serialized_json_user.return_value = serialized_user

    resp = callback(request)
    
    mock_serialized_json_user.assert_called_once_with("json", [User(id=1)])
    assert request.session.get('user') == serialized_user

    assert resp.status_code == 302
    assert resp['location'] == 'http://testserver/admin/content/'

@mock.patch('oauthadmin.views.OAuth2Session')
@mock.patch('oauthadmin.views.app_setting')
def test_logout_redirects(app_setting, OAuth2Session, request_factory):
    request = request_factory.get('/')
    request.session = {'oauth_token': 'token'}
    app_setting.return_value = 'app-setting'
    resp = logout(request)
    assert resp.status_code == 302
    assert resp['Location'] == 'http://testserver/'

@mock.patch('oauthadmin.views.OAuth2Session')
@mock.patch('oauthadmin.views.app_setting')
def test_logout_destroys_session(app_setting, OAuth2Session, request_factory):
    request = request_factory.get('/')
    request.session = {'oauth_token': 'token'}
    app_setting.return_value = 'app-setting'
    assert request.session.get('oauth_token') == 'token'
    resp = logout(request)
    assert request.session.get('oauth_token') == None



from setuptools import setup
from os.path import dirname, join

try:
    with open(join(dirname(__file__), 'README.md')) as f:
        README = f.read()
except IOError:
    README = '<no description>'

install_requires = ['oauthlib','requests_oauthlib']

try:
    import importlib
except ImportError:
    install_requires.append('importlib')


setup(
    name='django-admin-oauth2',
    version='1.3.0',
    description='A django app that replaces the django admin authentication mechanism by deferring to an oauth2 provider',
    long_description_content_type=README,
    url='https://github.com/RealGeeks/django-admin-oauth2',
    author='Real Geeks LLC',
    author_email='andrew@realgeeks.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    license='MIT',
    install_requires=install_requires,
    packages=['oauthadmin'],
)



from django.conf import settings as global_settings

defaults = {
    # default values for all django-oauth2-admin settings
    "GET_USER": 'oauthadmin.stubs.get_user',
    "PING_INTERVAL": 300,
    "OAUTHADMIN_DEFAULT_NEXT_URL": "/admin/",
    "SCOPE": ['default'],
}

global_prefix = 'OAUTHADMIN_'

def app_setting(name):
    return getattr(global_settings, global_prefix+name, defaults.get(name))



from django.core import serializers
from time import time
from oauthadmin.utils import import_by_path
from oauthadmin.settings import app_setting
from oauthadmin.views import destroy_session

try:
    # https://docs.djangoproject.com/en/1.10/topics/http/middleware/#upgrading-middleware
    from django.utils.deprecation import MiddlewareMixin
    BaseClass = MiddlewareMixin
except ImportError:
    BaseClass = object

def _ping_timeout_expired(timestamp, ping_interval):
    return time() - timestamp > ping_interval

def _verify_ping_interval(request, ping_interval, ping_func):
    if not _ping_timeout_expired(
        ping_interval,
        request.session.get('last_verified_at', 0),
    ):
        return
    request.session['last_verified_at'] = int(time())
    is_valid = ping_func(request.session['oauth_token'])
    if not is_valid:
        destroy_session(request)
        from django.contrib.auth.models import AnonymousUser
        request.user = AnonymousUser()

class OauthAdminSessionMiddleware(BaseClass):
    def process_request(self, request):
        if hasattr(request, 'session') and 'user' in request.session:
            request.user = list(serializers.deserialize("json", request.session['user']))[0].object
            request._cached_user = list(serializers.deserialize("json", request.session['user']))[0].object

            if app_setting('PING_INTERVAL') and app_setting('PING'):
                _verify_ping_interval(
                    request,
                    app_setting('PING_INTERVAL'),
                    import_by_path(app_setting('PING'))
                )

        else:
            from django.contrib.auth.models import AnonymousUser
            request.user = AnonymousUser()
